/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "Allocation.h"

#include "Register.h"

#define INFINITY 99999999

Temporary::Temporary(temporary_t type, int value)
{
  if(type == REGISTER)
    {
      is_register = true;
      is_stack = false;
      reg = value;
    }
  else //STACK
    {
      is_register = false;
      is_stack = true;
      shift = value;
    }       
}


int
Allocation::only_precolored_nodes(BitVector* nodes)
{
  //Termination case for Chaitin's algorithm
  vector<int> node_vec = nodes->to_vector();
  for(int i=0; i<node_vec.size(); i++)
    if(assignment.find(node_vec[i]) == assignment.end())
      return false;
  return true;
}


int 
Allocation::degree(InterferenceGraph* g, BitVector* nodes, int s)
{
  int result;
  vector<int> successors = g->succ[s]->to_vector();
  for(int i=0; i<successors.size(); i++)
    if(nodes->has(successors[i]))
      result++;
  return result;
}


int 
Allocation::remaining_color(InterferenceGraph* g, BitVector* nodes, int s, 
			    int K)
{
  BitVector* available_colors = new BitVector(K);
  for(int i=0; i<K; i++)
    available_colors->insert(i);

  vector<int> successors = g->succ[s]->to_vector();
  for(int i=0; i<successors.size(); i++)
    if(nodes->has(successors[i]))
      //For each successor
      if(assignment[successors[i]]->is_register)
	//... remove its color from the set.
	available_colors->remove(assignment[successors[i]]->reg);

  for(int i=0; i<K; i++)
    if(available_colors->has(i))
      return i; 
  //return the first available color (affinity handling should be there)
	
  return -1; //no available color.
}


void
Allocation::chaitin(InterferenceGraph* g, BitVector* nodes, int K)
{
  //if(nodes->is_empty())
  //  return;

  if(only_precolored_nodes(nodes))
    return;

  //
  // If there exists a node s trivially colorable, 
  // Then: 
  // - color(G - {s})
  // - assign to s a remaining register
  //

  int exists_trivially_colorable_node = false;
  for(int i=0; i<nb_registers(); i++)
    if(nodes->has(i))
      //For each remaining node ...
      if(assignment.find(i) == assignment.end())
	//... not precolored...
      if(degree(g,nodes,i) < K)
	{ //node trivially colorable?
	  nodes->remove(i); //remove the node
	  chaitin(g,nodes,K); //color G - {node}
	  nodes->insert(i); //restore the node

	  assignment[i] = new Temporary(REGISTER,remaining_color(g,nodes,i,K));

	  exists_trivially_colorable_node = true;
	  break;
	}

  //
  // Else: 
  // - Pick a node s with max degree 
  // - color(G - {s})
  // - assign a remaining color to s if exists (otherwise, spill s)
  //

  if(!exists_trivially_colorable_node)
    {
      int max_degree = -1;
      int s;
      for(int i=0; i<nb_registers(); i++)
	if(nodes->has(i))
	  if(assignment.find(i) == assignment.end()) //not precolored
	  {
	    int d = degree(g,nodes,i);
	    if(d > max_degree)
	      {
		max_degree = d;
		s = i;
	      }
	  }
      //s = node with a maximum degree

      nodes->remove(s); //remove the node
      chaitin(g,nodes,K); //color G - {node}
      nodes->insert(s); //restore the node

      int color = remaining_color(g,nodes,s,K);
      if(color != -1)
	//color found?
	assignment[s] = new Temporary(REGISTER,color); 
      else
	//no color found => spill
	assignment[s] = new Temporary(STACK,0); 
    }
}


Allocation::Allocation(InterferenceGraph* g, int K, map<int,int> precolor)
{
  BitVector* nodes = new BitVector(nb_registers());
  for(int i=0; i<nb_registers(); i++)
    nodes->insert(i);
  //Initially, all the nodes of g exists

  for(map<int,int>::const_iterator it = precolor.begin(); it != precolor.end();
      ++it)
    assignment[it->first] = new Temporary(REGISTER,it->second);
  //Apply the precoloration

  /*
   * First round: allocate the registers
   */

  chaitin(g,nodes,K);

  /*
   * Second round: allocate the stack temporaries
   */

  BitVector* spill_nodes = new BitVector(nb_registers());
  vector<map<int,Temporary*>::iterator> to_erase;
  for(map<int,Temporary*>::iterator it = assignment.begin(); 
      it != assignment.end(); ++it)
    if(it->second->is_stack)
      {
	spill_nodes->insert(it->first);
	to_erase.push_back(it);
      }
  //Restrict the interference graph to the spill nodes

  for(int i=0; i<to_erase.size(); i++)
    assignment.erase(to_erase[i]);
  //Remove the spills from the current allocation 
  //(would be considered precolored otherwise)

  vector<int> spill_vec = spill_nodes->to_vector();

  chaitin(g,spill_nodes,256);      
  //Allocate with MEMORY_SIZE positions

  nb_temporary = -1;
  for(int i=0; i<spill_vec.size(); i++)
    {
      int shift = assignment[spill_vec[i]]->reg;
      assignment[spill_vec[i]] = new Temporary(STACK,-1*shift);
      if(shift>nb_temporary) nb_temporary = shift;
    }
  nb_temporary++;
  //Set shift to register # 

}


//Pretty printer
ostream& operator <<(ostream& sout, Temporary& p)
{
  if(p.is_register)
    sout << "r" << p.reg;
  else
    sout << "ARP + " << p.shift;
  return sout;
}


ostream& operator <<(ostream& sout, Allocation& p)
{
  sout << "Stack size: " << p.nb_temporary << endl;
  for(map<int,Temporary*>::const_iterator it = p.assignment.begin(); 
      it != p.assignment.end(); ++it)
    sout << it->first << " --> " << *(it->second) << endl;
  return sout;
}
