/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "Backend.h"

#include "BasicBlock.h"
#include "Dag.h"
#include "InterferenceGraph.h"
#include "Allocation.h"

#include "Register.h"
#include "Label.h"

#include <stdlib.h>
#include <fstream> //for dag output
#include <iostream>
#include <fstream> //for .asm output


extern ofstream asm_output;

Cfg* digmips_cfg;

vector<int> marked; //DAG nodes already traversed
vector<int> result_reg;

int select_root(Dag* dag, int node_rank, BitVector* live_out)
{
  if(marked[node_rank])
    return result_reg[node_rank];

  marked[node_rank] = true;

  /*
   * First, produce the code for each successor
   */
  
  vector<int> operand_register;
  for(int i=0; i<dag->succ[node_rank].size(); i++)
    {
      int reg = select_root(dag,dag->succ[node_rank][i],live_out);
      operand_register.push_back(reg);
    }


  /*
   * Now, produce the code for the current node
   */

  pseudo_code_t instruction = dag->node[node_rank]->instruction;
  int result_register;
  if(dag->roots.find(node_rank) != dag->roots.end())
    //current node is a root of the DAG?
    result_register = dag->node[node_rank]->r_dest;
  else
    result_register = new_register();


  // Rewrite the node's pseudo-code to use son's registers
  // and result_register

  PseudoCode* to_translate; //node code with modified i/o registers

  if(instruction == ADD)
    to_translate = pseudo_add(result_register,operand_register[0],
			      operand_register[1]);

  if(instruction == SUB)
    to_translate = pseudo_sub(result_register,operand_register[0],
			      operand_register[1]);

  if(instruction == MUL)
    to_translate = pseudo_mul(result_register,operand_register[0],
			      operand_register[1]);

  if(instruction == DIV)
    to_translate = pseudo_div(result_register,operand_register[0],
			      operand_register[1]);

  if(instruction == MOV_REG_INT)
    to_translate = pseudo_mov_reg_int(result_register,
				      dag->node[node_rank]->constant);

  if(instruction == COPY)
    {
      to_translate = 0; //nothing to translate
      result_register = dag->node[node_rank]->r1; //it should be a DAG leaf
    }

  if(instruction == LOAD)
    to_translate = pseudo_load(result_register,operand_register[0],
			       dag->node[node_rank]->shift);

  if(instruction == STORE)
    to_translate = pseudo_store(operand_register[1],
				dag->node[node_rank]->shift,
				operand_register[0]);

  if(instruction == JUMP)
    to_translate = dag->node[node_rank];

  if(instruction == CJUMP) 
    to_translate = pseudo_cjump(operand_register[0], 
				dag->node[node_rank]->relation,
				operand_register[1],
				dag->node[node_rank]->label_id);

  if(instruction == CALL)
    to_translate = dag->node[node_rank];

  if(instruction == RET)
    to_translate = dag->node[node_rank];

  if(instruction == LABEL)
    to_translate = dag->node[node_rank];

  if(instruction == PUSH_ALL)
    to_translate = dag->node[node_rank];

  if(instruction == POP_ALL)
    to_translate = dag->node[node_rank];

  if(instruction == PUSH)
    to_translate = pseudo_push(operand_register[0]);

  if(instruction == POP)
    to_translate = pseudo_pop(result_register);

  if(instruction == LOAD_ARP)
    to_translate = pseudo_load_arp(result_register,
				   dag->node[node_rank]->shift);

  if(instruction == STORE_ARP)
    to_translate = pseudo_store_arp(dag->node[node_rank]->shift,
				    operand_register[0]);

  if(instruction == ADD_SP_REG)
    to_translate = pseudo_add_sp_reg(operand_register[0]);

  if(instruction == SUB_SP_REG)
    to_translate = pseudo_sub_sp_reg(operand_register[0]);

  if(instruction == MOV_REG_ARP)
    to_translate = pseudo_mov_reg_arp(result_register);

  if(instruction == MOV_REG_SP)
    to_translate = pseudo_mov_reg_sp(result_register);

  if(instruction == MOV_SP_REG)
    to_translate = pseudo_mov_sp_reg(operand_register[0]);

  if(instruction == PUSH_ARP)
    to_translate = dag->node[node_rank];

  if(instruction == POP_ARP)
    to_translate = dag->node[node_rank];

  if(instruction == MOV_ARP_SP)
    to_translate = dag->node[node_rank];

  if(instruction == MOV_SP_ARP)
    to_translate = dag->node[node_rank];

  if(instruction == INLINE_ASM)
    to_translate = dag->node[node_rank];

  // Produce the corresponding Digmips code.
  if(to_translate != 0)
    digmips_translate(digmips_cfg,to_translate);

  //
  // Copy the result into each live-out temporary
  //

  vector<int> def_registers = dag->node_def[dag->node[node_rank]];
  for(int i=0; i<def_registers.size(); i++)
    if(live_out->has(def_registers[i]) && 
       def_registers[i] != result_register)
      //def register for node_rank is live-out ?
      digmips_translate(digmips_cfg,
			pseudo_copy(def_registers[i],result_register)); 
        //live-out := result

  result_reg[node_rank] = result_register;

  return result_register;
}


void select_dag(Dag* dag, BitVector* live_out)
{
  marked.resize(dag->node.size());
  for(int i=0; i<dag->node.size(); i++)
    marked[i] = false;
  //Initially, no DAG node is marked

  result_reg.resize(dag->node.size());

  for(set<int>::const_iterator it = dag->roots.begin(); 
      it != dag->roots.end(); ++it)
    //For each root node...
    select_root(dag,*it,live_out); //produce the code
}


Cfg* select(Cfg* cfg)
{
  digmips_cfg = new Cfg(); //global CFG to bee filled.

  /*
   * Produce the code for each basic block.
   */

  for(int i=0; i<cfg->node.size(); i++)
    //For each basic block...
    {
      Dag* dag = new Dag((BasicBlock*)cfg->node[i]);
      //Get the DAG...

      //dag->simplify(cfg->live_out[i]);
      //Optimize the DAG
      
      select_dag(dag,cfg->live_out[i]);
      //Produce the code.        
    }

  digmips_cfg->fix_succ();
  digmips_cfg->do_liveness();

  return digmips_cfg;
}


void
backend(Cfg* cfg)
{
  // ofstream output_file;
  // output_file.open("dag.dot");
  // for(int i=0; i<cfg->node.size(); i++)
  //   {
  //     Dag* dag = new Dag((BasicBlock*)cfg->node[i]);
  //     dag->print_dot(output_file);
  //     //dag->simplify(cfg->live_out[i]);
  //     //dag->print_dot(output_file);
  //   }
  // output_file.close();
  // //Produce the .dot
  // system("dot -Tps dag.dot > dag.ps");
  // system("evince dag.ps &");
  // //Compile and display

  // cfg->show_dot();

  // cfg->print_code(cout);


  /*
   * 1) Selection and scheduling.
   */

  Cfg* digmips_virtual_cfg = select(cfg);
  //digmips_virtual_cfg->print_code();
  //digmips_virtual_cfg->show_dot();
  //digmips_virtual_cfg->extract_basic_blocks()->show_dot();


  /*
   * 2) Register & stack allocation
   */

  InterferenceGraph* interferences = new InterferenceGraph(digmips_virtual_cfg);

  //interferences->show_dot();

  map<int,int> precolor;
  precolor[2] = 0; //return_register

  Allocation* allocation = new Allocation(interferences,4,precolor);
  //Allocate among r2, r3, r4, r5 (actually r0, r1, r2, r3)

  for(map<int,Temporary*>::const_iterator it = allocation->assignment.begin();
      it != allocation->assignment.end(); ++it)
    if(it->second->is_register)
      it->second->reg += 2;
  //Shift the registers to allocate among r2,r3,r4,r5

  //cout << *allocation << endl;  

  /*
   * 3) Application to the code + spill code generation
   */

  Cfg* final_code = new Cfg();

  int prelude_patched = false;
  for(int i=0; i<digmips_virtual_cfg->node.size(); i++)
    {
      CodeDigmips* code = (CodeDigmips*)digmips_virtual_cfg->node[i];

      if((code->instruction == DIGMIPS_LDI) && (code->constant == 999) &&
	 !prelude_patched)
	{
	  prelude_patched = true;
	  code->constant = allocation->nb_temporary;
	}
      //Patch the prelude to allocate the right number of temporaries
	
      digmips_apply(final_code,code,allocation);
      //Apply the allocation (spill code generated here)
    }

  /*
   * 4) Emission of final code.
   */

  final_code->print_code(asm_output);
}
