/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "Label.h"
#include "Cfg.h"
#include "BasicBlock.h"

#include <stdlib.h> //for exit() and system()
#include <fstream> //for show_dot() file I/O

Cfg::Cfg() {}


void
Cfg::add_node(Code* node)
{
  int nb_node = this->node.size();

  if(nb_node >= 1)
    //if(!(node[nb_node-1]->is_jump()))
      this->succ[nb_node-1].push_back(nb_node); 
  //The new node will be a successor of the current one.

  this->node.push_back(node);
  //WARNING: no clonage

  vector<int> empty_vec;
  this->succ.push_back(empty_vec);

  if(node->is_label())
    label_node[node->get_label()] = nb_node;

  this->live_in.push_back(new BitVector(0));
  this->live_out.push_back(new BitVector(0));
}


void 
Cfg::set_live_in(BitVector* node_live_in)
{
  live_in[live_in.size()-1] = node_live_in->clone();
}


void 
Cfg::set_live_out(BitVector* node_live_out)
{
  live_out[live_out.size()-1] = node_live_out->clone();
}


void 
Cfg::fix_succ()
{
  for(int i=0; i<node.size(); i++)
    if(node[i]->is_jump())
      //For each jump/cjump node...
      if(label_node.find(node[i]->get_label()) != label_node.end())
	succ[i].push_back(label_node[node[i]->get_label()]);
        //Add the label to the successors
}


void 
Cfg::do_liveness()
{
  int fix_point_reached; //true if nothing has changed in the iteration

  do 
    {
      fix_point_reached = true;

      vector<BitVector*> new_live_in ( live_in.size() );
      vector<BitVector*> new_live_out ( live_out.size() );
      
      for(int i=0; i<node.size(); i++)
	//For each node...
	{
	  /*
	   * live_in(n) = use(n) U (prev_live_out(n) - def(n))
	   */
	  
	  BitVector& use = *(node[i]->get_read_register());
	  BitVector& def = *(node[i]->get_write_register());
	  new_live_in[i] = &(use | (*(live_out[i]) - def));

	  /*
	   * live_out(n) = U_{s \in succ(n)} prev_live_in(s)
	   */
	
	  new_live_out[i] = new BitVector(0);
	  for(int j=0; j<succ[i].size(); j++)
	    new_live_out[i] = &(*(new_live_out[i]) | *(live_in[succ[i][j]]));

	  /*
	   *  Something changed?
	   */

	  if(fix_point_reached)
	    if(!((*(new_live_in[i]) == *(live_in[i])) && 
		 (*(new_live_out[i]) == *(live_out[i]))))
	      //yes => continue the iterations
	      fix_point_reached = false;
	}

      if(!fix_point_reached)
	for(int i=0; i<node.size(); i++)
	  {
	    live_in[i] = new_live_in[i];
	    live_out[i] = new_live_out[i];
	  }
      //live_in = new_live_in, live_out = new_live_out
    }
  while(!fix_point_reached);
}


Cfg*
Cfg::extract_basic_blocks()
{
  Cfg* result = new Cfg(); //empty CFG

  vector<Code*> current_block;
  BitVector* current_live_in;

  map<string,int> bb; //bb[label] = #BB starting with 'label'
  vector<int> jump;      //jump = #BB finishing with a jump
  vector<string> jump_label; //corresponding jump label

  for(int i=0; i<node.size(); i++)
    //For each node...
    {
      if(node[i]->is_jump() &&
	 label_node.find(node[i]->get_label()) != label_node.end())	
	//Jump (inside the function)? => will end current block
	{
	  if(current_block.size() == 0)
	    current_live_in = live_in[i];
	  current_block.push_back(node[i]); 
	  result->add_node(new BasicBlock(current_block)); 
	  result->set_live_in(current_live_in);
	  result->set_live_out(live_out[i]);
	  //add jump to current block

	  jump.push_back(result->node.size()-1);
	  jump_label.push_back(node[i]->get_label());
	  current_block.resize(0);
	  continue;
	}

      if(node[i]->is_label())
	//Label? => will start next block
	{
	  if(current_block.size() != 0)
	    //close current block
	    {
	      result->add_node(new BasicBlock(current_block));
	      result->set_live_in(current_live_in);
	      result->set_live_out(live_out[i-1]);
	    }
	  current_block.resize(0);
	  current_block.push_back(node[i]); 
	  //start next block with the label
	  current_live_in = live_in[i];
	  //Set live_in to live_in(label)

	  bb[node[i]->get_label()] = result->node.size();
	  continue;
	}

      if(current_block.size() == 0)
	//Start a block? => set live_in
	//(i=0, or just after a jump)
	current_live_in = live_in[i];
      current_block.push_back(node[i]);
    }

  if(current_block.size() != 0)
    {
      result->add_node(new BasicBlock(current_block));
      result->set_live_in(current_live_in);
      result->set_live_out(live_out[live_out.size()-1]);
      //not a jump-finishing BB => liveness not set
    }


  //Fix jumps
  for(int i=0; i<jump.size(); i++)
    result->succ[jump[i]].push_back(bb[jump_label[i]]);

  return result;
}


void
Cfg::print_dot(ostream& sout)
{
  sout << "digraph g {"<< endl;

  //Nodes
  for(int i=0; i<node.size(); i++)
    sout << "  node" << i << "[label=\"" 
	 << "live-in: " << live_in[i]->to_string_dot() << "\\n"
	 << node[i]->to_string_dot()
	 << "\\nlive-out: " << live_out[i]->to_string_dot() << "\\n"
	 << "\",shape=box];" << endl;

  //Edges
  for(int i=0; i<succ.size(); i++)
    for(int j=0; j<succ[i].size(); j++)
      sout << "  node" << i << " -> node" << succ[i][j] << endl;

  sout << "}";
}


void 
Cfg::show_dot()
{
  ofstream output_file;
  output_file.open("cfg.dot");
  print_dot(output_file);
  output_file.close();
  //Produce the .dot

  system("dot -Tps cfg.dot > cfg.ps");
  system("evince cfg.ps &");
  //Compile and display
}


void 
Cfg::print_code(ostream& sout)
{
  for(int i=0; i<node.size(); i++)
    sout << node[i]->to_string() << endl;
}


Cfg* 
Cfg::clone()
{
  Cfg* result = new Cfg();
  
  for(int i=0; i<node.size(); i++)
    result->node.push_back(node[i]->clone());
  //Nodes

  result->succ.resize(succ.size());
  for(int i=0; i<succ.size(); i++)
    for(int j=0; j<succ[i].size(); j++)
      result->succ[i].push_back(succ[i][j]);
  //Successors

  for(map<string,int>::const_iterator it = label_node.begin(); 
      it != label_node.end(); ++it)
    result->label_node[(*it).first] = (*it).second;
  //Label nodes #
  
  for(int i=0; i<live_in.size(); i++)
    result->live_in.push_back(live_in[i]->clone());
  //Live-in

  for(int i=0; i<live_out.size(); i++)
    result->live_out.push_back(live_out[i]->clone());
  //Live-out
}


//pretty-printer
ostream& operator <<(ostream& sout, Cfg& g)
{
  for(int i=0; i<g.node.size(); i++)
    {
      sout << "Node #" << i << endl
	   << "  " << g.node[i]->to_string() << endl
	   << "  Successors: {";
      for(int j=0; j<g.succ[i].size(); j++)
	sout << g.succ[i][j] << " ";
      sout << "}" << endl
	   << "  Live-in: " << *(g.live_in[i]) << endl
	   << "  Live-out: " << *(g.live_out[i]) << endl
	   << endl;
    }

  return sout;
}

//Pretty constructors

extern Cfg* cfg; //global CFG to be filled

void add(int r_dest, int r1, int r2)
{
  cfg->add_node(pseudo_add(r_dest,r1,r2));
}

void sub(int r_dest, int r1, int r2)
{
  cfg->add_node(pseudo_sub(r_dest,r1,r2));
}

void mul(int r_dest, int r1, int r2)
{
  cfg->add_node(pseudo_mul(r_dest,r1,r2));
}

void div(int r_dest, int r1, int r2)
{
  cfg->add_node(pseudo_div(r_dest,r1,r2));
}

void mov_reg_int(int r_dest, int constante)
{
  cfg->add_node(pseudo_mov_reg_int(r_dest,constante));
}

void copy(int r_dest, int r_source)
{
  cfg->add_node(pseudo_copy(r_dest,r_source));
}

void load(int r_dest, int r_base, int shift)
{
  cfg->add_node(pseudo_load(r_dest,r_base,shift));
}

void store(int r_dest, int shift, int r_source)
{
  cfg->add_node(pseudo_store(r_dest,shift,r_source));
}

void jump(string label)
{
  cfg->add_node(pseudo_jump(label));
}

void cjump(int r1, test_t relation, int r2, string label)
{
  cfg->add_node(pseudo_cjump(r1,relation,r2,label));
}

void call(string label)
{
  cfg->add_node(pseudo_call(label));
}

void ret(int nb_args)
{
  cfg->add_node(pseudo_ret(nb_args));
}

void label(string label)
{
  cfg->add_node(pseudo_label(label));
}

void push_all()
{
  cfg->add_node(pseudo_push_all());
}

void pop_all()
{
  cfg->add_node(pseudo_pop_all());
}

void push(int reg)
{
  cfg->add_node(pseudo_push(reg));
}

void pop(int reg)
{
  cfg->add_node(pseudo_pop(reg));
}

void load_arp(int reg, int shift)
{
  cfg->add_node(pseudo_load_arp(reg,shift));
}

void store_arp(int shift, int reg)
{
  cfg->add_node(pseudo_store_arp(shift,reg));
}

void add_sp_reg(int reg)
{
  cfg->add_node(pseudo_add_sp_reg(reg));
}

void sub_sp_reg(int reg)
{
  cfg->add_node(pseudo_sub_sp_reg(reg));
}

void mov_reg_arp(int reg)
{
  cfg->add_node(pseudo_mov_reg_arp(reg));
}

void mov_reg_sp(int reg)
{
  cfg->add_node(pseudo_mov_reg_sp(reg));
}

void mov_sp_reg(int reg)
{
  cfg->add_node(pseudo_mov_sp_reg(reg));
}

void push_arp()
{
  cfg->add_node(pseudo_push_arp());
}

void pop_arp()
{
  cfg->add_node(pseudo_pop_arp());
}

void mov_arp_sp()
{
  cfg->add_node(pseudo_mov_arp_sp());
}

void mov_sp_arp()
{
  cfg->add_node(pseudo_mov_sp_arp());
}

void inline_asm(string asm_code)
{
  cfg->add_node(pseudo_inline_asm(asm_code));
}

