/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#include "PseudoCode.h"

#include <stdlib.h> //for exit()
#include <sstream> //for to_string()

#define RETURN_REGISTER 2

PseudoCode::PseudoCode(pseudo_code_t op, int r_dest, int r1, int r2)
{
  instruction = op;
  this->r_dest = r_dest;
  this->r1 = r1;

  read_register = new BitVector(0);
  write_register = new BitVector(0);

  if(op == LOAD || op == STORE)
    {
      this->shift = r2; //then r2 is the (constant) shift 

      read_register->insert(r1);
      if(op == LOAD)
	//r_dest = [r1+shift] 
	// => read = {r1}, write = {r_dest}
	write_register->insert(r_dest);
      else
	//[r_dest+shift] = r1 
	// => read = {r1,r_dest}, write = {}
	read_register->insert(r_dest);
    }  
  else //ADD/SUB/MUL/DIV
    {
      this->r2 = r2;

      read_register->insert(r1);
      read_register->insert(r2);
      write_register->insert(r_dest);
      //r_dest = r1 op r2 => read = {r1,r2}, write = {r_dest}
    }
}
 

PseudoCode::PseudoCode(pseudo_code_t op, int r_dest, int r1)
{
  instruction = op;
  this->r_dest = r_dest;

  read_register = new BitVector(0);
  write_register = new BitVector(0);

  if(op == LOAD_ARP)
    {
      this->shift = r1;
      write_register->insert(r_dest);
      return;
    }

  if(op == STORE_ARP)
    {
      this->r1 = r1;
      this->shift = r_dest;
      read_register->insert(r1);
      return;
    }

  if(op == MOV_REG_INT)
    this->constant = r1;
  else //COPY
    {
      this->r1 = r1;
      read_register->insert(r1);
      //r_dest = r1
    }
  write_register->insert(r_dest);
}


PseudoCode::PseudoCode(pseudo_code_t op, string label)
{
  instruction = op;
  this->label_id = label;

  read_register = new BitVector(0);
  write_register = new BitVector(0);
  if(op == CALL)
    write_register->insert(4); //function call --> r4 is written  
}


PseudoCode::PseudoCode(pseudo_code_t op, int r1, test_t relation, int r2, 
		       string label)
{
  instruction = op;
  this->relation = relation;
  this->r1 = r1;
  this->r2 = r2;
  this->label_id = label;

  read_register = new BitVector(0);
  write_register = new BitVector(0);

  read_register->insert(r1);
  read_register->insert(r2);
}


PseudoCode::PseudoCode(pseudo_code_t op, int nb_args)
{
  instruction = op;
  read_register = new BitVector(0);
  write_register = new BitVector(0);

  if(op == RET)
    {
      constant = nb_args;
      read_register->insert(RETURN_REGISTER); //should be alive 'til RET.
    }

  if(op == PUSH || op == ADD_SP_REG || op == SUB_SP_REG || op == MOV_SP_REG)
    {
      r1 = nb_args;
      read_register->insert(r1);
    }

  if(op == POP || op == MOV_REG_ARP || op == MOV_REG_SP)
    {
      r_dest = nb_args;
      write_register->insert(r_dest);
    }
}


PseudoCode::PseudoCode(pseudo_code_t op)
{
  instruction = op;
  read_register = new BitVector(0);
  write_register = new BitVector(0);
}


BitVector* 
PseudoCode::get_read_register()
{
  return read_register;
}


BitVector* 
PseudoCode::get_write_register()
{
  return write_register;
}


string
PseudoCode::get_label()
{
  return label_id;
}


int 
PseudoCode::is_jump()
{
  return instruction == JUMP || instruction == CJUMP;
}


int 
PseudoCode::is_label()
{
  return instruction == LABEL;
}


PseudoCode* 
PseudoCode::clone()
{
  PseudoCode* result = new PseudoCode(COPY,r_dest,r1); //Dummy
  
  result->instruction = instruction;

  result->r_dest = r_dest;
  result->r1 = r1;
  result->r2 = r2;

  result->shift = shift; 

  result->relation = relation;
  result->label_id = label_id; 
  result->asm_code = asm_code; 

  result->read_register = read_register->clone();
  result->write_register = write_register->clone();

  return result;
}


int
PseudoCode::is_leaf()
{
  return ((instruction == JUMP)
	  || (instruction == CALL)
	  || (instruction == RET)
	  || (instruction == LABEL)
	  || (instruction == PUSH_ALL)
	  || (instruction == POP_ALL)
	  || (instruction == POP)
	  || (instruction == LOAD_ARP)
	  || (instruction == MOV_REG_ARP)
	  || (instruction == MOV_REG_SP)
	  || (instruction == PUSH_ARP)
	  || (instruction == POP_ARP)
	  || (instruction == MOV_ARP_SP)
	  || (instruction == MOV_SP_ARP)
	  || (instruction == INLINE_ASM));
}


int
PseudoCode::is_ldi()
{
  return instruction == MOV_REG_INT;
}


int
PseudoCode::is_copy()
{
  return instruction == COPY;
}


int
PseudoCode::is_unary()
{
  return ((instruction == LOAD)
	  || (instruction == PUSH)
	  || (instruction == STORE_ARP)
	  || (instruction == ADD_SP_REG)
	  || (instruction == SUB_SP_REG)
	  || (instruction == MOV_SP_REG));
}


int
PseudoCode::is_binary()
{
  return ((instruction == ADD)
	  || (instruction == SUB)
	  || (instruction == MUL)
	  || (instruction == DIV)
	  || (instruction == STORE)
	  || (instruction == CJUMP));
}


int 
PseudoCode::is_write()
{
  return ((instruction == ADD) 
	  ||(instruction == SUB)
	  ||(instruction == MUL)
	  ||(instruction == DIV)
	  ||(instruction == MOV_REG_INT)
	  ||(instruction == COPY)
	  ||(instruction == LOAD)
	  ||(instruction == POP)
	  ||(instruction == LOAD_ARP)
	  ||(instruction == MOV_REG_ARP)
	  ||(instruction == MOV_REG_SP));
}


int 
PseudoCode::has_side_effect()
{
  return ((instruction == LOAD)
	  ||(instruction == STORE)
	  ||(instruction == CALL) 
	  ||(instruction == RET)
	  ||(instruction == PUSH)
	  ||(instruction == POP)
	  ||(instruction == LOAD_ARP)
	  ||(instruction == STORE_ARP)
	  ||(instruction == ADD_SP_REG)
	  ||(instruction == SUB_SP_REG)
	  ||(instruction == MOV_REG_ARP)
	  ||(instruction == MOV_REG_SP)
	  ||(instruction == MOV_SP_REG)
	  ||(instruction == PUSH_ARP)
	  ||(instruction == POP_ARP)
	  ||(instruction == MOV_SP_ARP)
	  ||(instruction == MOV_ARP_SP)
	  ||(instruction == INLINE_ASM));
}


//pretty-printer
string 
PseudoCode::to_string()
{
  string result;
  stringstream sout;

  /*
    sout << "Read = " << *read_register 
    << "\tWrite = " << *write_register
    << "\t";
  */

  if(instruction == LABEL)
    sout << label_id << ":";
  else
    sout << "  "; //Indentation

  if(instruction == ADD)
    sout << "r" << r_dest << " = r" << r1 << " + r" << r2; 
  
  if(instruction == SUB)
    sout << "r" << r_dest << " = r" << r1 << " - r" << r2; 

  if(instruction == MUL)
    sout << "r" << r_dest << " = r" << r1 << " * r" << r2; 

  if(instruction == DIV)
    sout << "r" << r_dest << " = r" << r1 << " / r" << r2; 

  if(instruction == MOV_REG_INT)
    sout << "r" << r_dest << " = " << constant;

  if(instruction == COPY)
    sout << "r" << r_dest << " = r" << r1;

  if(instruction == LOAD)
    sout << "r" << r_dest << " = [r" << r1 << " + " << shift << "]";

  if(instruction == STORE)
    sout << "[r" << r_dest << " + " << shift << "] = r" << r1;
  
  if(instruction == JUMP)
    sout << "jump " << label_id;
  
  if(instruction == CJUMP)
    {
      sout << "cjump r" << r1 << " ";
      if(relation == E) 
	sout << "==";
      if(relation == NE) 
	sout << "!=";
      if(relation == L) 
	sout << "<";
      if(relation == LE) 
	sout << "<=";
      if(relation == G) 
	sout << ">";
      if(relation == GE) 
	sout << ">=";
      sout << " r" << r2 << " --> " << label_id;
    }
  
  if(instruction == CALL)
    sout << "call " << label_id;
  
  if(instruction == RET)
    sout << "ret " << constant;
  
  if(instruction == PUSH_ALL)
    sout << "push_all";
  
  if(instruction == POP_ALL)
    sout << "pop_all";
  
  if(instruction == PUSH)
    sout << "push r" << r1;
  
  if(instruction == POP)
    sout << "pop r" << r_dest;
  
  if(instruction == LOAD_ARP)
    sout << "r" << r_dest << " = [ARP + " << shift << "]";
  
  if(instruction == STORE_ARP)
    sout << "[ARP + " << shift << "] = r" << r1;
  
  if(instruction == ADD_SP_REG)
    sout << "SP = SP + r" << r1;
  
  if(instruction == SUB_SP_REG)
    sout << "SP = SP - r" << r1;
  
  if(instruction == MOV_REG_ARP)
    sout << "r" << r_dest << " = ARP";
  
  if(instruction == MOV_REG_SP)
    sout << "r" << r_dest << " = SP";
  
  if(instruction == MOV_SP_REG)
    sout << "SP = r" <<  r1;
  
  if(instruction == PUSH_ARP)
    sout << "push ARP";
  
  if(instruction == POP_ARP)
    sout << "pop ARP";
  
  if(instruction == MOV_ARP_SP)
    sout << "ARP = SP";

  if(instruction == MOV_SP_ARP)
    sout << "SP = ARP";

  if(instruction == INLINE_ASM)
    sout << "-- " << asm_code << "";
  
  result = sout.str(); //convert back to string
  return result;
}


string 
PseudoCode::to_string_dot()
{
  return to_string();
}


PseudoCode* pseudo_add(int r_dest, int r1, int r2)
{
  return new PseudoCode(ADD,r_dest,r1,r2);
}

PseudoCode* pseudo_sub(int r_dest, int r1, int r2)
{
  return new PseudoCode(SUB,r_dest,r1,r2);
}

PseudoCode* pseudo_mul(int r_dest, int r1, int r2)
{
  return new PseudoCode(MUL,r_dest,r1,r2);
}

PseudoCode* pseudo_div(int r_dest, int r1, int r2)
{
  return new PseudoCode(DIV,r_dest,r1,r2);
}

PseudoCode* pseudo_mov_reg_int(int r_dest, int constant)
{
  return new PseudoCode(MOV_REG_INT,r_dest,constant);
}

PseudoCode* pseudo_copy(int r_dest, int r_source)
{
  return new PseudoCode(COPY,r_dest,r_source);
}

PseudoCode* pseudo_load(int r_dest, int r_base, int shift)
{
  return new PseudoCode(LOAD,r_dest,r_base,shift);
}

PseudoCode* pseudo_store(int r_dest, int shift, int r_source)
{
  return new PseudoCode(STORE,r_dest,r_source,shift);
}

PseudoCode* pseudo_jump(string label)
{
  return new PseudoCode(JUMP,label);
}

PseudoCode* pseudo_cjump(int r1, test_t relation, int r2, string label)
{
  return new PseudoCode(CJUMP,r1,relation,r2,label);
}

PseudoCode* pseudo_call(string label)
{
  return new PseudoCode(CALL,label);
}

PseudoCode* pseudo_ret(int nb_args)
{
  return new PseudoCode(RET,nb_args);
}

PseudoCode* pseudo_label(string label)
{
  return new PseudoCode(LABEL,label);
}

PseudoCode* pseudo_push_all()
{
  return new PseudoCode(PUSH_ALL);
}

PseudoCode* pseudo_pop_all()
{
  return new PseudoCode(POP_ALL);
}

PseudoCode* pseudo_push(int reg)
{
  return new PseudoCode(PUSH,reg);
}

PseudoCode* pseudo_pop(int reg)
{
  return new PseudoCode(POP,reg);
}

PseudoCode* pseudo_load_arp(int reg, int shift)
{
  return new PseudoCode(LOAD_ARP,reg,shift);
}

PseudoCode* pseudo_store_arp(int shift, int reg)
{
  return new PseudoCode(STORE_ARP,shift,reg);
}

PseudoCode* pseudo_add_sp_reg(int reg)
{
  return new PseudoCode(ADD_SP_REG,reg);
}

PseudoCode* pseudo_sub_sp_reg(int reg)
{
  return new PseudoCode(SUB_SP_REG,reg);
}

PseudoCode* pseudo_mov_reg_arp(int reg)
{
  return new PseudoCode(MOV_REG_ARP,reg);
}

PseudoCode* pseudo_mov_reg_sp(int reg)
{
  return new PseudoCode(MOV_REG_SP,reg);
}

PseudoCode* pseudo_mov_sp_reg(int reg)
{
  return new PseudoCode(MOV_SP_REG,reg);
}

PseudoCode* pseudo_push_arp()
{
  return new PseudoCode(PUSH_ARP);
}

PseudoCode* pseudo_pop_arp()
{
  return new PseudoCode(POP_ARP);
}

PseudoCode* pseudo_mov_arp_sp()
{
  return new PseudoCode(MOV_ARP_SP);
}

PseudoCode* pseudo_mov_sp_arp()
{
  return new PseudoCode(MOV_SP_ARP);
}

PseudoCode* pseudo_inline_asm(string asm_code)
{
  PseudoCode* result = new PseudoCode(INLINE_ASM); //no read, no write.
  result->asm_code = asm_code;
  return result;    
}


/*
void emit(PseudoCode* code)
{
  cout + *code +  endl;
}
*/
