/*****************************************************************************
 *                DigCC: a minimal C compiler for DigMIPS                    *
 ***************************************************************************** 
 *                                                                           *
 * Copyright (c) 2012 Christophe Alias                                       *
 *                                                                           *
 * This program is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU General Public License as published by the     *
 * Free Software Foundation; either version 2 of the License, or (at your    * 
 * option) any later version.                                                *
 *                                                                           *
 * This program is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General  *
 * Public License for more details.                                          *
 *                                                                           *
 * You should have received a copy of the GNU General Public License along   *
 * with this program; if not, write to the Free Software Foundation, Inc.,   *
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                 *
 *                                                                           *
 * Written by Christophe Alias, Christophe.Alias@ens-lyon.fr                 *
 *                                                                           *
 *****************************************************************************/
#ifndef __PSEUDO_CODE_H
#define __PSEUDO_CODE_H

#include "Code.h" //abstract super-class

#include "BitVector.h"

#include <iostream>  //For the pretty-printer
#include <vector>

using namespace std;

enum pseudo_code_t
  {
    ADD, SUB, MUL, DIV, //Arithmetic operations
    MOV_REG_INT, COPY, 
    LOAD, STORE, 
    JUMP, CJUMP, 
    CALL, RET,   // Function call
    LABEL,
    
    PUSH_ALL, POP_ALL,
    PUSH, POP,
    LOAD_ARP,    //
    STORE_ARP,   // Argument access

    ADD_SP_REG,  //
    SUB_SP_REG,  //
    MOV_REG_ARP, // Stack allocation
    MOV_REG_SP,  //
    MOV_SP_REG,  //

    PUSH_ARP,     //
    POP_ARP,      // Function prelude/postlude
    MOV_ARP_SP,   //
    MOV_SP_ARP,   //

    INLINE_ASM
  };

enum test_t
  {
    E, NE, L, LE, G, GE
  };

class PseudoCode: public Code
{
 public:
  pseudo_code_t instruction;

  int r_dest;
  int r1;
  int r2;

  int constant; //MOV_REG_INT

  int shift; //LOAD/STORE

  test_t relation; // CJUMP
  string label_id;  // Jumps / label

  string asm_code; //inline asm

  BitVector* read_register; //use
  BitVector* write_register; //def

  PseudoCode(pseudo_code_t op, int r_dest, int r1, int r2);
  //ADD, SUB, MUL, DIV
  //LOAD/STORE --> (pseudo_code_t op, int r_dest, int r1, int shift);

  PseudoCode(pseudo_code_t op, int r_dest, int r1);
  //MOV_REG_INT, COPY
  //LOAD_ARP, STORE_ARP

  PseudoCode(pseudo_code_t op, string label);
  //JUMP, CALL, LABEL

  PseudoCode(pseudo_code_t op, int r1, test_t relation, int r2, string label);
  //CJUMP

  PseudoCode(pseudo_code_t op, int nb_args);
  //RET, PUSH, POP, ADD_SP_REG, SUB_SP_REG, MOV_REG_ARP, MOV_REG_SP, MOV_SP_REG.

  PseudoCode(pseudo_code_t op);
  //PUSH_ALL, POP_ALL, PUSH_BP, POP_BP, MOV_BP_SP, MOV_ARP_SP, MOV_SP_ARP

  PseudoCode* clone();

  //
  // Implementation of abstract methods of Code
  //

  BitVector* get_read_register(); //use
  BitVector* get_write_register(); //def
  string get_label();

  int is_jump();
  int is_label();

  string to_string();
  string to_string_dot();


  //further methods (used in DAG construction)
  int is_leaf(); //0 read
  int is_unary(); //1 read
  int is_binary(); //2 reads
  int is_ldi();
  int is_copy();

  int is_write(); //write r_dest?
  int has_side_effect(); 
  //hidden write (to SP, ARP and stack)?
  //DAG construction will avoid any simplification, 
  //then code generation will preserve there execution order.
};

//Pretty constructors
PseudoCode* pseudo_add(int r_dest, int r1, int r2);
PseudoCode* pseudo_sub(int r_dest, int r1, int r2);
PseudoCode* pseudo_mul(int r_dest, int r1, int r2);
PseudoCode* pseudo_div(int r_dest, int r1, int r2);
PseudoCode* pseudo_mov_reg_int(int r_dest, int constante);
PseudoCode* pseudo_copy(int r_dest, int r_source);
PseudoCode* pseudo_load(int r_dest, int r_base, int shift);
PseudoCode* pseudo_store(int r_dest, int shift, int r_source);
PseudoCode* pseudo_jump(string label);
PseudoCode* pseudo_cjump(int r1, test_t relation, int r2, string label);
PseudoCode* pseudo_call(string label);
PseudoCode* pseudo_ret(int nb_args);
PseudoCode* pseudo_label(string label);

PseudoCode* pseudo_push_all();
PseudoCode* pseudo_pop_all();
PseudoCode* pseudo_push(int reg);
PseudoCode* pseudo_pop(int reg);
PseudoCode* pseudo_load_arp(int reg, int shift);
PseudoCode* pseudo_store_arp(int shift, int reg);
PseudoCode* pseudo_add_sp_reg(int reg);
PseudoCode* pseudo_sub_sp_reg(int reg);
PseudoCode* pseudo_mov_reg_arp(int reg);
PseudoCode* pseudo_mov_reg_sp(int reg);
PseudoCode* pseudo_mov_sp_reg(int reg);
PseudoCode* pseudo_push_arp();
PseudoCode* pseudo_pop_arp();
PseudoCode* pseudo_mov_arp_sp();
PseudoCode* pseudo_mov_sp_arp();

PseudoCode* pseudo_inline_asm(string asm_code);

#endif
